"""
Utility functions to ease path manipulations.
The API manipulates paths for either files or directories:

    D:/dir1/dir2       -> dir2      is a file      (no ending /)
    D:/dir1/file.txt/  -> file.txt/ is a directory (   ending /)

Converts antislashes to normal slashes.
"""

import glob
import os.path
import sys

this_module = sys.modules[__name__]

class _join_class:
    def __call__(self, arg1, *args): # type: (str, *str) -> str
        """ /dir1/dir2   file.txt -> /dir1/dir2/file.txt
            /dir1/dir2/  dir3/    -> /dir1/dir2/dir3/
        """
        return no_antislash(os.path.join(arg1, *args))
    def __getitem__(self, args): # type: (*str) -> str
        """ /dir1/dir2   file.txt -> /dir1/dir2/file.txt/
            /dir1/dir2/  dir3/    -> /dir1/dir2/dir3/
        """
        if isinstance(args, tuple):
            return slash(join.__call__(*args))
        return slash(join.__call__(args))

join = _join_class()

def add(arg1, *args): # type: (str, *str) -> str
    """ /dir1/dir2   file.txt -> /dir1/dir2file.txt
        /dir1/dir2/  dir3/    -> /dir1/dir2/dir3/
    """
    return no_antislash("".join([arg1, *args]))

def files(path = "", pattern = "*"): # type: (str, str) -> list(str)
    """ /dir1/dir2/  *.txt -> [/dir1/dir2/file.txt]
    """
    return [no_antislash(x) for x in glob.glob(dirpath(path) + pattern, recursive=True) if os.path.isfile(x)]

def dirs(path = "", pattern = "*"): # type: (str, str) -> list(str)
    """ /dir1/*/ -> [/dir1/dir2/]
    """
    return [slash(x) for x in glob.glob(dirpath(path) + pattern, recursive=True) if os.path.isdir(x)]

def filepath(path): # type: (str) -> str
    """ /dir1/dir2/file.txt -> /dir1/dir2/file.txt
        /dir1/dir2/         -> /dir1/dir2/
    """
    return no_antislash(path)

def filename(path): # type: (str) -> str
    """ /dir1/dir2/file.txt -> file.txt
        /dir1/dir2/         ->
    """
    if not path:
        return ""
    if path[-1] in ['/', '\\']:
        return ""
    return no_antislash(os.path.basename(path))

def set_filename(path, filename): # type: (str, str) -> str
    """ /dir1/dir2/file.txt -> /dir1/dir2/newfile.xml
        /dir1/dir2/         -> /dir1/dir2/newfile.xml
    """
    return dirpath(path) + no_antislash(filename)

def dirpath(path): # type: (str) -> str
    """ /dir1/dir2/file.txt -> /dir1/dir2/
        /dir1/dir2/         -> /dir1/dir2/
    """
    if not path:
        return ""
    path = no_antislash(path)
    if path[-1] == '/':
        return path
    result = os.path.dirname(path)
    if not result:
        return ""
    if result[-1] == '/':
        return result
    return result + '/'

def dirname(path): # type: (str) -> str
    """ /dir1/dir2/file.txt -> dir2
        /dir1/dir2/         -> dir2
    """
    result = dirpath(path)
    if not result:
        return ""
    return result.rsplit('/', 2)[-2]

def parent_dirpath(path): # type: (str) -> str
    """ /dir1/dir2/file.txt -> /dir1/
        /dir1/dir2/         -> /dir1/
    """
    return dirpath(os.path.dirname(no_antislash(path)))

def parent_dirname(path): # type: (str) -> str
    """ /dir1/dir2/file.txt -> dir1
        /dir1/dir2/         -> dir1
    """
    return dirname(os.path.dirname(no_antislash(path)))

def ext(path): # type: (str) -> str
    """ /dir1/dir2/file.ext.txt -> .txt
        /dir1/dir2/             ->
    """
    return os.path.splitext(path)[1]

def set_ext(path, extension): # type: (str, str) -> str
    """ /dir1/dir2/file.ext.txt -> /dir1/dir2/file.ext.new
        /dir1/dir2/             -> /dir1/dir2/.new
    """
    return no_ext(path) + extension

def no_ext(path): # type: (str) -> str
    """ /dir1/dir2/file.ext.txt -> /dir1/dir2/file.ext
        /dir1/dir2/             -> /dir1/dir2/
    """
    return no_antislash(os.path.splitext(path)[0])

def fullext(path): # type: (str) -> str
    """ /dir1/dir2/file.ext.txt -> .ext.txt
        /dir1/dir2/             ->
    """
    if not path:
        return ""
    if path[-1] in ['/', '\\']:
        return ""
    splits = os.path.basename(path).split('.', 1)
    if len(splits) == 1:
        return ""
    return '.' + splits[1]

def set_fullext(path, extension): # type: (str, str) -> str
    """ /dir1/dir2/file.ext.txt -> /dir1/dir2/file.new
        /dir1/dir2/             -> /dir1/dir2/.new
    """
    return no_fullext(path) + extension

def no_fullext(path): # type: (str) -> str
    """ /dir1/dir2/file.ext.txt -> /dir1/dir2/file
        /dir1/dir2/             -> /dir1/dir2/
    """
    extension = fullext(path)
    if not extension:
        return no_antislash(path)
    return no_antislash(path[:-len(extension)])

def slash(path): # type: (str) -> str
    """ /dir1/dir2/file.txt -> /dir1/dir2/file.txt/
        /dir1/dir2/         -> /dir1/dir2/
    """
    if not path:
        return "/"
    path = no_antislash(path)
    if path[-1] == '/':
        return path
    return path + '/'

def no_slash(path): # type: (str) -> str
    if not path:
        return ""
    if path[-1] in ['/', '\\']:
        return no_antislash(path[:-1])
    return no_antislash(path)

def no_antislash(path): # type: (str) -> str
    """ \dir1\dir2\file.txt -> /dir1/dir2/file.txt/
        \dir1\dir2\         -> /dir1/dir2/
    """
    return path.replace('\\', '/')

def realpath(path): # type: (str) -> str
    if not path:
        return ""
    result = os.path.realpath(path)
    if path[-1] in ['/', '\\']:
        return slash(result)
    return no_antislash(result)

def relpath(path, start = "."): # type: (str, str) -> str
    if not path:
        return start
    result = os.path.relpath(path, start)
    if path[-1] in ['/', '\\']:
        return slash(result)
    return no_antislash(result)

def samefile(path1, path2): # type: (str, str) -> bool
    return os.path.samefile(path1, path2)

class Path(str):
    """ Allows a functional API of the above functions:

            path("/dir").join("file.txt").exists
                instead of
            os.path.exists(os.path.join("/dir", "file.txt"))
    """
    def __new__(cls, *args, **kwargs):
        if len(args) == 0:
            return str.__new__(cls, **kwargs)
        return str.__new__(cls, no_antislash(args[0]), *args[1:], **kwargs)
    def __getattribute__(self, attr):
        try:
            attribute = object.__getattribute__(self, attr)
            return PathWrapper(attribute) if callable(attribute) else attribute
        except:
            result = os.path.__getattribute__(attr)(self)
            return Path(result) if isinstance(result, str) else result

    class _join_class:
        def __init__(self, path):
            self.path = path
        def __call__(self, *args):
            path = join.__call__(self.path, *args)
            return Path(path)
        def __getitem__(self, args):
            if isinstance(args, tuple):
                path = slash(join.__call__(self.path, *args))
            else:
                path = slash(join.__call__(self.path, args))
            return Path(path)

    @property
    def join(self):
        return Path._join_class(self)
    def add(self, *args):
        path = add(self, *args)
        return Path(path)
    def __add__(self, other):
        return self.add(other)
    def __radd__(self, other):
        return Path(other.__add__(self))
    def __truediv__(self, other):
        return self.join(other)
    def files(self, pattern = "*"):
        list = files(self, pattern)
        return [Path(path) for path in list]
    def dirs(self, pattern = "*"):
        list = dirs(self, pattern)
        return [Path(path) for path in list]
    @property
    def filepath(self):
        path = filepath(self)
        return Path(path)
    @property
    def filename(self):
        path = filename(self)
        return Path(path)
    def set_filename(self, filename):
        path = set_filename(self, filename)
        return Path(path)
    @property
    def dirpath(self):
        path = dirpath(self)
        return Path(path)
    @property
    def dirname(self):
        path = dirname(self)
        return Path(path)
    @property
    def parent_dirpath(self):
        path = parent_dirpath(self)
        return Path(path)
    @property
    def parent_dirname(self):
        path = parent_dirname(self)
        return Path(path)
    @property
    def ext(self):
        path = ext(self)
        return Path(path)
    def set_ext(self, extension):
        path = set_ext(self, extension)
        return Path(path)
    @property
    def no_ext(self):
        path = no_ext(self)
        return Path(path)
    @property
    def fullext(self):
        path = fullext(self)
        return Path(path)
    def set_fullext(self, extension):
        path = set_fullext(self, extension)
        return Path(path)
    @property
    def no_fullext(self):
        path = no_fullext(self)
        return Path(path)
    @property
    def slash(self):
        path = slash(self)
        return Path(path)
    @property
    def no_slash(self):
        path = no_slash(self)
        return Path(path)
    @property
    def realpath(self):
        path = realpath(self)
        return Path(path)
    def relpath(self, start = "."):
        path = relpath(self, start)
        return Path(path)
    def samefile(self, path2):
        path = samefile(self, path2)
        return Path(path)

class PathWrapper:
    """ Helps forwarding to callables of os.path
    """
    def __init__(self, attribute):
        self.attribute = attribute
    def __call__(self, *args, **kwargs):
        result = self.attribute.__call__(*args, **kwargs)
        return Path(result) if isinstance(result, str) else result
    def __getitem__(self, args):
        result = self.attribute.__getitem__(args)
        return Path(result) if isinstance(result, str) else result
    def __getattr__(self, attr):
        return getattr(self.attribute, attr)

class module_call:
    """ Overrides path(), path[] and path.attr
    """
    def __call__(self, path):
        """ path("/my/path") -> "/my/path"
        """
        return Path(path)
    def __getitem__(self, path):
        """ path["/my/path"] -> "/my/path/"
        """
        return Path(slash(path))
    def __getattribute__(self, attr):
        """ Forwards os.path if not defined in this module
        """
        try:
            return this_module.__getattribute__(attr)
        except:
            return os.path.__getattribute__(attr)

sys.modules[__name__] = module_call()



from .defaults import *
from .exceptions import *

import io
import os

class ParserSettings:
    def __init__(self,
                 cog_open = AUTOJINJA_DEFAULT_COG_OPEN,
                 cog_close = AUTOJINJA_DEFAULT_COG_CLOSE,
                 cog_end = AUTOJINJA_DEFAULT_COG_END,
                 cog_as_comment = False,
                 edit_open = AUTOJINJA_DEFAULT_EDIT_OPEN,
                 edit_close = AUTOJINJA_DEFAULT_EDIT_CLOSE,
                 edit_end = AUTOJINJA_DEFAULT_EDIT_END,
                 edit_as_comment = False,
                 remove_markers = None,
                 encoding = None,
                 newline = None):
        self.cog_open = cog_open
        self.cog_close = cog_close
        self.cog_end = cog_end
        self.cog_as_comment = cog_as_comment
        self.edit_open = edit_open
        self.edit_close = edit_close
        self.edit_end = edit_end
        self.edit_as_comment = edit_as_comment
        self.remove_markers = remove_markers
        self.encoding = encoding
        self.newline = newline

    @property
    def remove_markers(self):
        return self._removeMarkers
    @remove_markers.setter
    def remove_markers(self, remove_markers):
        self._removeMarkers = remove_markers or osenviron_remove_markers()

class Parser:
    def __init__(self, string, settings):
        self.string = string
        self.settings = settings
        self.markers = None # List[Marker]
        self.blocks = None # List[Block]
        self.cog_blocks = None # List[CogBlock]
        self.edit_blocks = None # List[EditBlock]

    @property
    def edits(self):
        return { key: edit_block.body for key, edit_block in self.edit_blocks.items() }

    def parse(self):
        idx = 0
        cog_markers = []
        edit_markers = []
        ### Find all [[[ ]]] pairs
        while True:
            marker = Marker(self.string, False, self.settings.cog_open, self.settings.cog_close, self.settings.cog_end, self.settings.cog_as_comment, None, True)
            marker, idx = self.find_marker(marker, idx, len(self.string))
            if not marker:
                break
            cog_markers.append(marker)
        ### Find all <<[ ]>> pairs not in [[[ ]]]
        sections = [[0,-1]]
        for cog_marker in cog_markers:
            sections[-1][1] = cog_marker.header_open
            sections.append([cog_marker.header_close, -1])
        sections[-1][1] = len(self.string)
        for section in sections:
            idx = section[0]
            while True:
                marker = Marker(self.string, True, self.settings.edit_open, self.settings.edit_close, self.settings.edit_end, self.settings.edit_as_comment, True, False)
                marker, idx = self.find_marker(marker, idx, section[1])
                if not marker:
                    break
                edit_markers.append(marker)
        ### Properly resolve [[[ ]]] and <<[ ]>>
        self.markers = cog_markers + edit_markers
        self.markers.sort(key=lambda x: x.header_open)
        self.check_markers()
        ### Collect blocks
        self.blocks = []
        self.cog_blocks = []
        self.edit_blocks = {}
        for marker in self.markers:
            if not marker.is_end:
                if marker.is_edit:
                    block = EditBlock(marker)
                    if block.name in self.edit_blocks:
                        raise DuplicateEditException.from_marker(marker)
                    self.edit_blocks[block.name] = block
                else:
                    block = CogBlock(marker)
                    self.cog_blocks.append(block)
                self.blocks.append(block)

    def find_marker(self, marker, idx, end):
        ### Find open
        found, idx = self.find_token(marker.open, idx, end)
        if not found:
            return None, idx
        ### Set open location
        marker.header_open = idx
        ### Find close
        idx = idx + len(marker.open)
        found, idx = self.find_token(marker.close, idx, end)
        if not found:
            raise CloseMarkerNotFoundException.from_marker(marker)
        ### Set close location
        idx = idx + len(marker.close)
        marker.header_close = idx
        ### Extract header
        marker.extract_indent()
        marker.extract_header()
        return marker, idx

    def find_token(self, token, idx, end):
        i = self.string.find(token, idx, end)
        if i < 0:
            return False, idx
        return True, i

    def check_markers(self):
        stack = []
        prev_inline = []
        mkrstate = 0 # 0: none, 1: cog, 2: edit, 3: cog end, 4: edit end
        state = 0 # 0: no requirement, 1: waiting for requirement, 2: require other line, 3: require same line
        prev_marker = None
        for marker in self.markers:
            ### Check cog & edit markers enclosure
            newstate = 1 + marker.is_edit + 2 * marker.is_end
            if mkrstate == 1 and newstate == 4:
                raise WrongInclusionException.from_marker(marker)
            if mkrstate == 2 and newstate == 3:
                raise WrongInclusionException.from_marker(marker)
            mkrstate = newstate
            ### Check direct enclosure
            if not marker.is_end and not marker.direct_enclosure and stack and stack[-1].is_edit == marker.is_edit:
                raise DirectlyEnclosedEditException.from_marker(marker)
            ### Check requirements with previous marker
            if prev_marker:
                same_line = marker.same_line(prev_marker.header_close, marker.header_open)
                if state == 0:
                    pass
                elif state == 1:
                    state = 3 if same_line else 0
                elif state == 2:
                    if same_line:
                        raise RequireNewlineException.from_marker(marker)
                    state = 0
                elif state == 3:
                    if not same_line:
                        raise RequireInlineException.from_marker(marker)
                    state = 0
            ### Ensure marker subtree is correctly inlined
            if prev_inline:
                if prev_inline[-1] == None: # No requirement
                    prev_inline[-1] = same_line
                elif prev_inline[-1] == False: # Should be multiline
                    if same_line:
                        raise RequireNewlineException.from_marker(marker)
                elif not same_line: # Should be inlined
                    raise RequireInlineException.from_marker(marker)
            ### Deal marker
            if not marker.is_end:
                stack.append(marker)
                prev_inline.append(True if prev_inline and prev_inline[-1] else None)
                prev_marker = marker
            else:
                if not stack:
                    raise OpenMarkerNotFoundException.from_marker(marker)
                openMarker = stack.pop()
                same_line = prev_inline.pop()
                prev_marker = marker
                ### Set duals
                openMarker.dual = marker
                marker.dual = openMarker
                ### Extract body
                openMarker.extract_body(marker, same_line)
                if same_line:
                    state = 1
                    # Set start / end indexes
                    openMarker.header_start = openMarker.header_open
                    openMarker.header_end = openMarker.header_close
                    marker.header_start = marker.header_open
                    marker.header_end = marker.header_close
                else:
                    state = 2
                    # Set end indexes
                    openMarker.header_end = openMarker.header_close
                    i = self.string.find('\n', marker.header_close)
                    marker.header_end = i+1 if i >= 0 else len(self.string)
        if stack:
            raise EndMarkerNotFoundException.from_marker(stack[-1])

class Marker:
    def __init__(self, string, is_edit, open, close, end, as_comment, headerInline, directEnclosure):
        self.string = string
        self.is_edit = is_edit
        self.is_end = False
        self.open = open
        self.close = close
        self.end = end
        self.as_comment = as_comment # True or False
        self.direct_enclosure = directEnclosure # True or False
        # Header
        self.header_inline = headerInline # True: inline, False: Multiline, None: Both
        self.header_empty = True
        self.header_column = 0
        self.header = ""
        self._header_stripped = None
        self.header_start = 0 # Index, multiline / inline dependant
        self.header_end = 0 # Index, multiline / inline dependant
        self.header_open = 0 # Index, l-prolonged
        self.header_close = 0 # Index, r-prolonged
        self.header_indent_column = 0
        self.header_indent = ""
        # Body
        self.body_inline = False
        self.body_empty = True
        self.body_column = 0
        self.body = ""
        self._body_dedented = None
        self.body_start = 0 # Index, with indentation
        self.body_end = 0 # Index
        # Dual
        self.dual = None # Open or end marker

    @property
    def header_stripped(self):
        if self._header_stripped == None:
            self._header_stripped = self.header.strip()
        return self._header_stripped
    @property
    def body_dedented(self):
        if self._body_dedented == None:
            self._body_dedented = self.dedent_body()
        return self._body_dedented

    def same_line(self, startidx, idx):
        i = self.string.rfind('\n', startidx, idx)
        return i < 0

    def extract_indent(self):
        start = self.string.rfind('\n', 0, self.header_open)+1
        end = start
        body = self.header_open
        i = self.header_open
        while i > start:
            i -= 1
            if self.string[i] != ' ' and self.string[i] != '\t':
                end = i+1 # Comment end
                if not self.as_comment:
                    body = start
                    while i > start:
                        i -= 1
                        if self.string[i] == ' ' or self.string[i] == '\t':
                            body = i+1 # Comment start
                            break
                break
        self.header_column = self.header_open - start
        self.header_start = start
        self.header_indent_column = end - start
        self.header_indent = ''.join(['\t' if x == '\t' else ' ' for x in self.string[start:self.header_open]])
        self.body_column = body - start

    def extract_header(self):
        marker_content = io.StringIO()
        self.header_empty = True
        idx = self.header_close - len(self.close)
        start = self.header_open + len(self.open)
        ### Same line
        if self.same_line(self.header_open, idx):
            if self.header_inline == False:
                raise RequireHeaderMultilineException.from_marker(self)
            if self.string[start] == ' ':
                start += 1
            if self.string[idx-1] == ' ':
                idx -= 1
            marker_content.write(self.string[start:idx])
            self.header_empty = idx <= start
        ### Different lines
        else:
            if self.header_inline == True:
                raise RequireHeaderInlineException.from_marker(self)
            # First line
            i = self.string.find('\n', start, idx)
            if i > start:
                if self.string[start] == ' ':
                    start += 1
                marker_content.write(self.string[start:i])
                self.header_empty = False
            # Other lines
            while True:
                start = i+1
                i = self.string.find('\n', start, idx)
                # Middle lines
                if i >= 0:
                    self.check_indent(start, i)
                    if not self.header_empty:
                        marker_content.write('\n')
                    marker_content.write(self.string[start+self.header_column:i])
                    self.header_empty = False
                # Last line
                else:
                    self.check_indent(start, idx)
                    if idx-start > self.header_column:
                        if self.string[idx-1] == ' ':
                            idx -= 1
                        if not self.header_empty:
                            marker_content.write('\n')
                        marker_content.write(self.string[start+self.header_column:idx])
                        self.header_empty = False
                    break
        self.header = marker_content.getvalue()
        self.is_end = self.header == self.end

    def extract_body(self, end_marker, same_line):
        body = io.StringIO()
        ### Same line
        if same_line:
            self.body_inline = True
            i = self.string.find(' ', self.header_close, end_marker.header_open)
            if i < 0:
                self.body_start = self.header_close
            else:
                self.body_start = i+1
                self.header_close = i # Adjust marker (marker prolongement)
            i = self.string.rfind(' ', self.header_close, end_marker.header_open)
            if i < 0:
                self.body_end = end_marker.header_open
            else:
                self.body_end = i
                end_marker.header_open = i+1 # Adjust marker (marker prolongement)
            body.write(self.string[self.body_start:self.body_end])
        ### Different lines
        else:
            self.body_inline = False
            self.body_start = self.string.find('\n', self.header_close) + 1
            self.body_end = self.string.rfind('\n', self.header_close, end_marker.header_open) + 1
            self.header_close = self.body_start
            body.write(self.string[self.body_start:self.body_end-1])
        self.body = body.getvalue()
        self.body_empty = len(self.body) == 0

    def check_indent(self, start, end):
        idx = start + self.header_indent_column
        for i in range(start, start+self.header_column):
            if i < end:
                if self.string[i] == ' ':
                    if self.header_indent[i-start] != '\t': continue
                elif self.string[i] == '\t':
                    if self.header_indent[i-start] == '\t': continue
                elif i < idx: continue # Eat until comment end
            elif i >= idx and self.string[i] == '\n':
                break
            raise WrongHeaderIndentationException.from_marker(self, i)

    def dedent_body(self):
        if self.body_empty:
            return None
        if self.body_inline:
            return self.body
        return self.dedent(self.body)

    def dedent(self, output):
        result = io.StringIO()
        start = 0
        while True:
            idx = output.find('\n', start, len(output))
            mid = min(start + self.body_column, len(output) if idx < 0 else idx)
            for i in range(start, mid):
                if output[i] != ' ' and output[i] != '\t':
                    result.write(output[i:mid])
                    break
            if idx < 0:
                result.write(output[mid:])
                break
            else:
                end = idx+1
                result.write(output[mid:end])
                start = end
        return result.getvalue()

class Block:
    def __init__(self, marker):
        self.marker = marker

    @property
    def raw_header(self):
        return self.marker.header
    @property
    def header(self):
        return self.marker.header_stripped

    @property
    def raw_body(self):
        return self.marker.body
    @property
    def body(self):
        return self.marker.body_dedented

    def get_code(self, additional_lines = (0, 0)):
        if self.marker.body_inline:
            start = self.marker.header_start
            end = self.marker.dual.header_end
            # Complete line
            start = self.marker.string.rfind('\n', 0, start)+1
            i = self.marker.string.find('\n', end)
            if i >= 0:
                end = i+1
            else:
                end = len(self.marker.string)
        else:
            start = self.marker.header_start
            end = self.marker.dual.header_end
        for i in range(additional_lines[0]):
            if start == 0:
                break
            start = self.marker.string.rfind('\n', 0, start-1)+1
        for i in range(additional_lines[1]):
            if end > len(self.marker.string):
                break
            i = self.marker.string.find('\n', end+1)
            if i >= 0:
                end = i
            else:
                end = len(self.marker.string)
        if end > 0 and self.marker.string[end-1] == '\n':
            end -= 1
        return self.marker.string[start:end]

class CogBlock(Block):
    def __init__(self, marker):
        super().__init__(marker)

class EditBlock(Block):
    def __init__(self, marker):
        super().__init__(marker)
        self._body = None
        self.allow_code_loss = False # Use at your own risk
    def __str__(self):
        return f"{self.__repr__()} named {self.name}"

    @property
    def name(self):
        return self.header

    @property
    def body(self):
        return self._body or super().body
    @body.setter
    def body(self, body):
        self._body = body



from . import defaults
from . import exceptions
from . import path
from . import parser
from . import utils

import io
import jinja2
import os.path

class AutoLoader(jinja2.BaseLoader):
    """ Jinja2 loader to find templates near already loaded templates """
    all_dirpaths_used = set()

    def __init__(self, additional_dirpaths):
        if additional_dirpaths != None:
            for dirpath in additional_dirpaths:
                AutoLoader.all_dirpaths_used.add(dirpath)

    def get_source(self, environment, template):
        for dirpath in AutoLoader.all_dirpaths_used:
            filepath = dirpath.join(template)
            if filepath.exists:
                with open(filepath) as f:
                    source = f.read()
                mtime = os.path.getmtime(filepath)
                return source, filepath, lambda: mtime == os.path.getmtime(filepath)
        raise jinja2.TemplateNotFound(template)

class RawTemplate:
    """ Shared Jinja2 environment """
    environment = None

    @staticmethod
    def create_loader(additional_dirpaths = None):
        return AutoLoader(additional_dirpaths)

    @staticmethod
    def create_environment(*args, **kwargs):
        if "loader" not in kwargs:
            kwargs["loader"] = RawTemplate.create_loader()
        if "keep_trailing_newline" not in kwargs:
            kwargs["keep_trailing_newline"] = True
        if "lstrip_blocks" not in kwargs:
            kwargs["lstrip_blocks"] = True
        if "trim_blocks" not in kwargs:
            kwargs["trim_blocks"] = True
        if "undefined" not in kwargs:
            kwargs["undefined"] = jinja2.StrictUndefined
        return jinja2.Environment(*args, **kwargs)

    def __init__(self, string, input = None, output = None, encoding = None, newline = None, globals = None):
        if RawTemplate.environment == None:
            RawTemplate.environment = RawTemplate.create_environment()
        template = RawTemplate.environment.from_string(source = string, globals = globals)
        if input != None:
            template.filename = input
            AutoLoader.all_dirpaths_used.add(path(input).dirpath)
        else:
            template.filename = exceptions.format_text(string)
        self.jinja2_template = template
        self.string = string
        self.input = input
        self.output = output
        self.encoding = encoding
        self.newline = newline
    def __getattr__(self, attr):
        return getattr(self.template, attr)

    @staticmethod
    def from_file(input, output = None, encoding = None, newline = None, globals = None):
        try:
            with open(input, 'r', encoding = encoding or "utf-8") as file:
                return RawTemplate(file.read(), input, output, encoding, newline, globals)
        except BaseException as e:
            raise exceptions.clean_traceback(e) from None
    @staticmethod
    def from_string(string, output = None, encoding = None, newline = None, globals = None):
        try:
            return RawTemplate(string, None, output, encoding, newline, globals)
        except BaseException as e:
            raise exceptions.clean_traceback(e) from None

    def render_file(self, output = None, encoding = None, newline = None):
        try:
            return self.context().render_file(output, encoding, newline)
        except BaseException as e:
            raise exceptions.clean_traceback(e) from None
    def render(self):
        try:
            return self.context().render()
        except BaseException as e:
            raise exceptions.clean_traceback(e) from None

    def context(__autojinja_self__, *args, **kwargs):
        if "self" in kwargs:
            kwargs["this"] = kwargs["self"] # Avoid conflict with Jinja2
            del kwargs["self"]
        return RawTemplate.Context(__autojinja_self__, args, kwargs)

    class Context:
        """ Forwards Jinja2 data model """
        def __init__(self, template, args = (), kwargs = {}):
            self.template = template
            self.args = args
            self.kwargs = kwargs
        def __getattr__(self, attr):
            return getattr(self.template, attr)

        def render_file(self, output = None, encoding = None, newline = None):
            try:
                output = output or self.output
                assert output != None, "output filepath parameter can't be None"
                if defaults.osenviron_debug():
                    args, kwargs = utils.wrap_objects(*self.args, **self.kwargs)
                else:
                    args, kwargs = self.args, self.kwargs
                result = self.jinja2_template.render(*args, **kwargs)
                utils.generate_file(output, result, None, encoding or self.encoding, newline or self.newline)
                return result
            except BaseException as e:
                raise exceptions.clean_traceback(e) from None
        def render(self):
            try:
                if defaults.osenviron_debug():
                    args, kwargs = utils.wrap_objects(*self.args, **self.kwargs)
                else:
                    args, kwargs = self.args, self.kwargs
                return self.jinja2_template.render(*args, **kwargs)
            except BaseException as e:
                raise exceptions.clean_traceback(e) from None

class BaseTemplate:
    def __init__(self, string, input = None, output = None, settings = None, remove_markers = None, encoding = None, newline = None, globals = None):
        if input != None:
            AutoLoader.all_dirpaths_used.add(path(input).dirpath)
        self.string = string
        self.input = input
        self.output = output
        self.settings = settings or parser.ParserSettings()
        self.remove_markers = remove_markers
        self.encoding = encoding
        self.newline = newline
        self.globals = globals
        self.overriden_edits = None
        self.parser = None

    @property
    def remove_markers(self):
        return self._removeMarkers or self.settings.remove_markers
    @remove_markers.setter
    def remove_markers(self, remove_markers):
        self._removeMarkers = remove_markers

    @property
    def encoding(self):
        return self._encoding or self.settings.encoding
    @encoding.setter
    def encoding(self, encoding):
        self._encoding = encoding

    @property
    def newline(self):
        return self._newline or self.settings.newline
    @newline.setter
    def newline(self, newline):
        self._newline = newline

    @property
    def markers(self):
        return self.parser.markers
    @property
    def blocks(self):
        return self.parser.blocks
    @property
    def cog_blocks(self):
        return self.parser.cog_blocks
    @property
    def edit_blocks(self):
        return self.parser.edit_blocks

    @property
    def edits(self):
        if self.overriden_edits == None:
            return self.parser.edits
        return self.overriden_edits.copy()
    @edits.setter
    def edits(self, edits):
        self.overriden_edits = edits

    def render_file(self, output = None, remove_markers = None, encoding = None, newline = None):
        try:
            return self.context().render_file(output, remove_markers, encoding, newline)
        except BaseException as e:
            raise exceptions.clean_traceback(e) from None
    def render(self, output = None, remove_markers = None):
        try:
            return self.context().render(output, remove_markers)
        except BaseException as e:
            raise exceptions.clean_traceback(e) from None

    def context(__autojinja_self__, *args, **kwargs):
        if "self" in kwargs:
            kwargs["this"] = kwargs["self"] # Avoid conflict with Jinja2
            del kwargs["self"]
        return BaseTemplate.Context(__autojinja_self__, args, kwargs)

    class Context:
        """ Forwards Jinja2 data model """
        def __init__(self, template, args = (), kwargs = {}):
            self.template = template
            self.args = args
            self.kwargs = kwargs
        def __getattr__(self, attr):
            return getattr(self.template, attr)

        def render_file(self, output = None, remove_markers = None, encoding = None, newline = None):
            try:
                output = output or self.output
                assert output != None, "output filepath parameter can't be None"
                ### Retrieve output edits
                edit_blocks_to_generate = {} # Dict[str, EditBlock], for generation
                if not os.path.isfile(output): # File doesn't exist
                    old_content = None
                elif self.input and os.path.samefile(self.input, output): # Same file
                    old_content = self.string
                else: # Not same file
                    with open(output, 'r', encoding = encoding or self.encoding or "utf-8") as file:
                        old_content = file.read()
                        edit_blocks = utils.edit_blocks_from_string(old_content, self.settings)
                        edit_blocks_to_generate.update(edit_blocks)
                ### Render
                result = self.parser.generate(edit_blocks_to_generate, self.overriden_edits, remove_markers or self.remove_markers, self.globals, self.args, self.kwargs)
                utils.generate_file(output, result, old_content, encoding or self.encoding, newline or self.newline)
                return result
            except BaseException as e:
                raise exceptions.clean_traceback(e) from None
        def render(self, output = None, remove_markers = None):
            try:
                ### Retrieve output edits
                edit_blocks_to_generate = {} # Dict[str, EditBlock], for generation
                if output != None:
                    edit_blocks = utils.edit_blocks_from_string(output, self.settings)
                    edit_blocks_to_generate.update(edit_blocks)
                ### Render
                return self.parser.generate(edit_blocks_to_generate, self.overriden_edits, remove_markers or self.remove_markers, self.globals, self.args, self.kwargs)
            except BaseException as e:
                raise exceptions.clean_traceback(e) from None

class CogTemplate(BaseTemplate):
    def __init__(self, string, input = None, output = None, settings = None, remove_markers = None, encoding = None, newline = None, globals = None):
        super().__init__(string, input, output or input, settings, remove_markers, encoding, newline, globals)
        self.parser = CogGenerator(string, self.settings)
        self.parser.parse()

    @staticmethod
    def from_file(input, output = None, settings = None, remove_markers = None, encoding = None, newline = None, globals = None):
        try:
            with open(input, 'r', encoding = encoding or (settings.encoding if settings else None) or "utf-8") as file:
                return CogTemplate(file.read(), input, output, settings, remove_markers, encoding, newline, globals)
        except BaseException as e:
            raise exceptions.clean_traceback(e) from None
    @staticmethod
    def from_string(string, output = None, settings = None, remove_markers = None, encoding = None, newline = None, globals = None):
        try:
            return CogTemplate(string, None, output, settings, remove_markers, encoding, newline, globals)
        except BaseException as e:
            raise exceptions.clean_traceback(e) from None

class JinjaTemplate(BaseTemplate):
    def __init__(self, string, input = None, output = None, settings = None, remove_markers = None, encoding = None, newline = None, globals = None):
        super().__init__(string, input, output, settings, remove_markers, encoding, newline, globals)
        self.parser = JinjaGenerator(string, self.settings)
        self.parser.parse()

    @staticmethod
    def from_file(input, output = None, settings = None, remove_markers = None, encoding = None, newline = None, globals = None):
        try:
            with open(input, 'r', encoding = encoding or (settings.encoding if settings else None) or "utf-8") as file:
                return JinjaTemplate(file.read(), input, output, settings, remove_markers, encoding, newline, globals)
        except BaseException as e:
            raise exceptions.clean_traceback(e) from None
    @staticmethod
    def from_string(string, output = None, settings = None, remove_markers = None, encoding = None, newline = None, globals = None):
        try:
            return JinjaTemplate(string, None, output, settings, remove_markers, encoding, newline, globals)
        except BaseException as e:
            raise exceptions.clean_traceback(e) from None

### Generators

class BaseGenerator(parser.Parser):
    def __init__(self, string, settings):
        super().__init__(string, settings)

    def generate(self, edit_blocks_to_generate, overriden_edits, remove_markers, globals, args, kwargs):
        ### Save settings
        old_remove_markers = self.settings.remove_markers
        if self.settings.remove_markers != remove_markers:
            self.settings.remove_markers = remove_markers
        try:
            ### Generate
            self.edit_blocks_to_generate = self.edit_blocks.copy()
            self.overriden_edits = overriden_edits
            self.edit_blocks_generated = set()
            self.globals = globals
            self.args = args
            self.kwargs = kwargs
            output = self.generate_output(edit_blocks_to_generate) # To inherit
            ### Check unused edits
            diff = set(self.edit_blocks_to_generate) - self.edit_blocks_generated
            for name in diff:
                block = self.edit_blocks_to_generate[name]
                if not block.allow_code_loss:
                    raise exceptions.NonGeneratedEditException.from_marker(block.marker)
        finally:
            ### Restore settings
            if self.settings.remove_markers != old_remove_markers:
                self.settings.remove_markers = old_remove_markers
        return output

    def evaluate(self, string):
        generator = CogGenerator(string, self.settings)
        generator.parse()
        generator.edit_blocks_to_generate = generator.edit_blocks.copy()
        generator.edit_blocks_to_generate.update(self.edit_blocks_to_generate)
        generator.overriden_edits = self.overriden_edits
        generator.edit_blocks_generated = self.edit_blocks_generated
        generator.globals = self.globals
        generator.args = self.args
        generator.kwargs = self.kwargs
        return generator.generate_output({})

class CogGenerator(BaseGenerator):
    def __init__(self, string, settings):
        super().__init__(string, settings)

    def generate_output(self, edit_blocks_to_generate):
        self.edit_blocks_to_generate.update(edit_blocks_to_generate) # Update for generation
        stringio = io.StringIO()
        idx = 0
        depth = 0
        for marker in self.markers:
            ### End marker
            if marker.is_end:
                depth -= 1
                if depth == 0:
                    if not self.settings.remove_markers:
                        ### Write header
                        stringio.write(self.string[marker.header_start:marker.header_end])
                    idx = marker.header_end
            ### Open marker
            else:
                depth += 1
                if depth == 1:
                    stringio.write(self.string[idx:marker.header_start])
                    if not self.settings.remove_markers:
                        ### Write header
                        stringio.write(self.string[marker.header_start:marker.header_end])
                    ### Generate
                    output = self.generate_marker(marker)
                    ### Same line
                    if marker.body_inline:
                        if output == None:
                            output = ""
                        elif '\n' in output:
                            raise exceptions.RequireBodyInlineException.from_marker(marker)
                        if not self.settings.remove_markers:
                            stringio.write(f" {output} ")
                        else:
                            stringio.write(output)
                    ### Different lines
                    elif output != None:
                        bodyIndent = marker.header_indent[:marker.body_column]
                        output = output.replace('\n', f"\n{bodyIndent}")
                        stringio.write(f"{bodyIndent}{output}\n")
        ### End of file
        stringio.write(self.string[idx:])
        return stringio.getvalue()

    def generate_marker(self, marker):
        if not marker.is_edit:
            if marker.header_empty:
                return None
            ### Generate raw template
            template = RawTemplate(marker.header, globals = self.globals)
            output = template.context(*self.args, **self.kwargs).render()
        else:
            ### Check if edit already used
            key = marker.header_stripped
            if key in self.edit_blocks_generated:
                raise exceptions.AlreadyGeneratedEditException.from_marker(marker)
            self.edit_blocks_generated.add(key)
            ### Get edit body
            if self.overriden_edits != None and key in self.overriden_edits:
                output = self.overriden_edits[key]
            else:
                edit_block = self.edit_blocks_to_generate.get(key)
                output = edit_block.body if edit_block else None
            if output == None:
                return None
        if len(output) == 0:
            return ""
        try:
            ### Parse and generate again
            result = self.evaluate(output)
            if result.endswith('\n'):
                result = result[:-1]
            if self.settings.remove_markers and len(result) == 0:
                return None
            return result
        except BaseException as e:
            raise exceptions.CommonException.from_exception(e, marker) from None

class JinjaGenerator(BaseGenerator):
    def __init__(self, string, settings):
        super().__init__(string, settings)

    def generate_output(self, edit_blocks_to_generate):
        to_reinsert = {} # str -> [Marker, Marker]
        stringio = io.StringIO()
        idx = 0
        depth = 0
        prev_list = None
        for marker in self.markers:
            ### End marker
            if marker.is_end:
                if not marker.is_edit:
                    depth -= 1
                    if depth == 0:
                        prev_list[1] = marker # Update previous reinsert
                        idx = marker.header_end
                else:
                    if depth == 0:
                        stringio.write(self.string[idx:marker.header_end])
                        idx = marker.header_end
            ### Open marker
            else:
                if not marker.is_edit:
                    depth += 1
                    if depth == 1:
                        prev_list = [marker, None]
                        id = self.unique_id(idx)
                        to_reinsert[id] = prev_list # Update reinserts
                        stringio.write(self.string[idx:marker.header_start])
                        stringio.write(id) # Write id for later reinsertion
                else:
                    if depth == 0:
                        del self.edit_blocks_to_generate[marker.header_stripped] # Update for generation
        ### End of file
        stringio.write(self.string[idx:])
        return self.generate_reinsert(stringio.getvalue(), to_reinsert, edit_blocks_to_generate)

    def generate_reinsert(self, string, to_reinsert, edit_blocks_to_generate):
        ### Generate raw template
        template = RawTemplate(string, globals = self.globals)
        output = template.context(*self.args, **self.kwargs).render()
        ### Reinsert cog markers
        for id, (marker_start, marker_end) in to_reinsert.items():
            if not marker_start.is_edit:
                content = self.string[marker_start.header_start:marker_end.header_end]
                output = output.replace(id, content)
        ### Parse and generate again
        self.edit_blocks_to_generate.update(edit_blocks_to_generate) # Update for generation
        if len(output) > 0:
            output = self.evaluate(output)
        return output

    def unique_id(self, idx):
        return f"°#@[AUTOJINJA_{idx}]+&*"



import os

AUTOJINJA_DEFAULT_FILENAME   = "__jinja__.py"
AUTOJINJA_DEFAULT_TAG        = "autojinja"
AUTOJINJA_DEFAULT_COG_OPEN   = "[[["
AUTOJINJA_DEFAULT_COG_CLOSE  = "]]]"
AUTOJINJA_DEFAULT_COG_END    = "end"
AUTOJINJA_DEFAULT_EDIT_OPEN  = "<<["
AUTOJINJA_DEFAULT_EDIT_CLOSE = "]>>"
AUTOJINJA_DEFAULT_EDIT_END   = "end"

AUTOJINJA_DEBUG          = "AUTOJINJA_DEBUG"
AUTOJINJA_REMOVE_MARKERS = "AUTOJINJA_REMOVE_MARKERS"
AUTOJINJA_SILENT         = "AUTOJINJA_SILENT"
AUTOJINJA_SUMMARY        = "AUTOJINJA_SUMMARY"
AUTOJINJA_THIS_DIRPATH   = "THIS_DIRPATH"

def osenviron_debug(env = None):
    env = env or os.environ
    if AUTOJINJA_DEBUG not in env:
        return 0
    value = env[AUTOJINJA_DEBUG]
    is_valid = True
    if len(value) != 1:
        is_valid = False
    elif value != "0" and value != "1":
        is_valid = False
    if not is_valid:
        raise Exception(f"Expected 0 or 1 for environment variable '{AUTOJINJA_DEBUG}'")
    return 1 if value == "1" else 0

def osenviron_remove_markers(env = None):
    env = env or os.environ
    if AUTOJINJA_REMOVE_MARKERS not in env:
        return 0
    value = env[AUTOJINJA_REMOVE_MARKERS]
    is_valid = True
    if len(value) != 1:
        is_valid = False
    elif value != "0" and value != "1":
        is_valid = False
    if not is_valid:
        raise Exception(f"Expected 0 or 1 for environment variable '{AUTOJINJA_REMOVE_MARKERS}'")
    return 1 if value == "1" else 0

def osenviron_silent(env = None):
    env = env or os.environ
    if AUTOJINJA_SILENT not in env:
        return 0
    value = env[AUTOJINJA_SILENT]
    is_valid = True
    if len(value) != 1:
        is_valid = False
    elif value != "0" and value != "1":
        is_valid = False
    if not is_valid:
        raise Exception(f"Expected 0 or 1 for environment variable '{AUTOJINJA_SILENT}'")
    return 1 if value == "1" else 0

def osenviron_summary(env = None):
    env = env or os.environ
    if AUTOJINJA_SUMMARY not in env:
        return "1"
    value = env[AUTOJINJA_SUMMARY]
    is_valid = True
    if len(value) != 1 and len(value) != 3:
        is_valid = False
    else:
        for c in value:
            if c != "0" and c != "1":
                is_valid = False
                break
    if not is_valid:
        raise Exception(f"Expected 0, 1 or flags for environment variable '{AUTOJINJA_SUMMARY}'")
    return value

def osenviron_this_dirpath(env = None):
    env = env or os.environ
    return env.get(AUTOJINJA_THIS_DIRPATH)



import os.path
import setuptools

THIS_DIR = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(THIS_DIR, "README.md")) as file:
    README = file.read()

setuptools.setup(
    name="autojinja",
    version="1.8.1",
    description="Content generation with Jinja templates in between comments",
    author="Florian Popek",
    author_email="florian.popek@gmail.com",
    url="https://github.com/ldflo/autojinja",
    license="BSD",
    long_description_content_type="text/markdown",
    long_description=README,
    platforms="OS Independent",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Code Generators",
    ],
    packages=["autojinja"],
    install_requires=["jinja2"],
    entry_points={"console_scripts": ["autojinja=autojinja.__main__:entry_point"]},
)



__version__ = "1.0.0"

from . import main

import sys

def entry_point():
    try:
        main(*sys.argv[1:])
        sys.exit(0)
    except Exception as e:
        sys.stderr.write(f"{e}\n")
        sys.exit(1)

if __name__ == "__main__":
    entry_point()



import io
import os
import traceback

def wrap_callable(callable):
    """ Wraps the given callable in order to bypass AttributeError
        handling of Jinja and provide better exception stacktraces.
    """
    def autojinja_wrapped_fcall(*args, **kwargs):
        try:
            result = callable(*args, **kwargs)
            return wrap_object_attributes(result)
        except AttributeError as e:
            message = clean_wrapped_stacktrace(e)
            raise DebugAttributeError(message) from None
        except BaseException as e:
            message = clean_wrapped_stacktrace(e)
            raise wrap_exception(e, message).with_traceback(None)
    return autojinja_wrapped_fcall

### Base exception

class CommonException(Exception):
    @staticmethod
    def from_marker(exceptionType, marker, pos, size, message):
        line = line_at_index(marker.string, pos)
        (l, c) = index_to_coordinates(marker.string, pos)
        message = f"{message}\n{line}\n{' ' * (c-1)}{'^' * size} line {l}, column {c}"
        exception = exceptionType(message)
        exception.coordinates = (l, c)
        return exception
    @staticmethod
    def from_exception(exception, marker):
        exception = prepend_jinja2_traceback(exception)
        text = format_text(marker.header)
        (l, c) = index_to_coordinates(marker.string, marker.header_open)
        stack = f"\n  During {'reinsertion' if marker.is_edit else 'generation'} of \"{marker.open} {text} {marker.close}\" at line {l}, column {c}"
        message = str(exception).lstrip('\n')
        message = f"{stack}\n{message}"
        exception = wrap_exception(exception, message)
        exception.coordinates = (l, c)
        return exception

    def __init__(self, message):
        super().__init__(message)

### Parsing exception

class ParsingException(CommonException):
    def __init__(self, message):
        super().__init__(message)

class OpenMarkerNotFoundException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(OpenMarkerNotFoundException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Couldn't find corresponding open marker \"{marker.open} {marker.close}\":")
    def __init__(self, message):
        super().__init__(message)

class CloseMarkerNotFoundException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(CloseMarkerNotFoundException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Couldn't find corresponding close marker \"{marker.close}\":")
    def __init__(self, message):
        super().__init__(message)

class EndMarkerNotFoundException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(EndMarkerNotFoundException,
                                           marker,
                                           marker.header_close - len(marker.close),
                                           len(marker.close),
                                           f"Couldn't find corresponding end marker \"{marker.open} {marker.end} {marker.close}\":")
    def __init__(self, message):
        super().__init__(message)

class RequireHeaderInlineException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(RequireHeaderInlineException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Marker can't have a multiline header:")
    def __init__(self, message):
        super().__init__(message)

class RequireHeaderMultilineException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(RequireHeaderMultilineException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Marker can't have a one line header:")
    def __init__(self, message):
        super().__init__(message)

class WrongHeaderIndentationException(ParsingException):
    @staticmethod
    def from_marker(marker, pos):
        return CommonException.from_marker(WrongHeaderIndentationException,
                                           marker,
                                           pos,
                                           len(marker.open),
                                           f"Wrong marker header indentation:")
    def __init__(self, message):
        super().__init__(message)

class RequireNewlineException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(RequireNewlineException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Marker can't start on same line as previous end marker:")
    def __init__(self, message):
        super().__init__(message)

class RequireInlineException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(RequireInlineException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Marker must start on same line as previous marker:")
    def __init__(self, message):
        super().__init__(message)

class WrongInclusionException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(WrongInclusionException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Marker has wrong inclusion regarding enclosing markers:")
    def __init__(self, message):
        super().__init__(message)

class DuplicateEditException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(DuplicateEditException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Duplicate edit marker \"{marker.open} {marker.header} {marker.close}\", consider reusing/removing duplicates:")
    def __init__(self, message):
        super().__init__(message)

class DirectlyEnclosedEditException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(DirectlyEnclosedEditException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Directly enclosed edit marker \"{marker.open} {marker.header} {marker.close}\", consider reusing/removing it:")
    def __init__(self, message):
        super().__init__(message)

### Generation exception

class GenerationException(CommonException):
    def __init__(self, message):
        super().__init__(message)

class RequireBodyInlineException(GenerationException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(RequireBodyInlineException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Generated body must contain only one line to be inlined:")
    def __init__(self, message):
        super().__init__(message)

class NonGeneratedEditException(GenerationException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(NonGeneratedEditException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Non-generated edit marker \"{marker.open} {marker.header} {marker.close}\", consider reusing/removing it:")
    def __init__(self, message):
        super().__init__(message)

class AlreadyGeneratedEditException(GenerationException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(AlreadyGeneratedEditException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Already generated edit marker \"{marker.open} {marker.header} {marker.close}\", consider fixing generation:")
    def __init__(self, message):
        super().__init__(message)

### Utils

def index_to_coordinates(string, index):
    """ Returns the corresponding tuple (line, column) of the character at the given index of the given string.
    """
    if index < 0:
        index =  index % len(string)
    sp = string[:index+1].splitlines(keepends=True)
    return len(sp), len(sp[-1])

def line_at_index(string, index):
    """ Returns the line at the given index of the given string.
    """
    if index < 0:
        index =  index % len(string)
    start = string.rfind('\n', 0, index)+1
    end = string.find('\n', index)
    suffix = '\\n' if end >= 0 else '\\0'
    if end < 0:
        end = len(string)
    return f"{string[start:end]}{suffix}"

def format_text(text, max_length = 50, suffix = "..."):
    """ Replaces tabulations and ends of line.
    """
    text = text if len(text) <= max_length else f"{text[:max_length]}{suffix}"
    return text.replace('\t', "\\t").replace('\n', "\\n")

def split_traceback(tb, remove_class_name = False):
    """ Splits the given traceback into (stacktrace, message).
    """
    startidx = -1
    while True:
        startidx = tb.find('\n', startidx+1)
        if startidx < 0:
            break # No stacktrace
        if not tb.startswith('  ', startidx+1):
            break # End of stacktrace
    if startidx < 0:
        stacktrace = None # No stacktrace
    else:
        stacktrace = tb[:startidx]
    if not remove_class_name:
        message = tb[startidx+1:]
    else:
        idx = tb.find(' ', startidx+1)
        if idx < 0:
            message = None # No message
        else:
            idx += 2 if tb[idx+1] == '\n' else 1 # lstrip one
            message = tb[idx:]
    return stacktrace, message

_disallowed_attrs = [
    "__delattr__",
    "__getattr__",
    "__getattribute__",
    "__init__",
    "__new__",
    "__setattr__",
    "__str__",
]

def wrap_exception(exception, message):
    """ Wraps the given exception with a custom message.
        Creates an object that mocks the exception class.
    """
    if hasattr(exception, "_wrapped_exception") and hasattr(exception, "_wrapped_message"):
        exception._wrapped_message = message
        return exception
    try:
        def __init__(self):
            pass
        def __str__(self):
            return self._wrapped_message
        dicts = { name: getattr(exception, name) for name in dir(exception)
                  if name.startswith("__") and name not in _disallowed_attrs }
        dicts["__init__"] = __init__
        dicts["__module__"] = exception.__class__.__module__
        dicts["__str__"] = __str__
        mock = type(exception.__class__.__qualname__, (exception.__class__,), dicts)()
        for key, value in mock.__dict__.items():
            setattr(mock, key, value)
        mock._wrapped_exception = exception
        mock._wrapped_message = message
        return mock
    except:
        return exception

def prepend_jinja2_traceback(exception, tb = None):
    """ Prepends the Jinja2 traceback to the given exception.
        Creates an object that derives the exception class.
    """
    if tb == None:
        tb = traceback.format_exc().rstrip('\n')
    (stacktrace, message) = split_traceback(tb, True)
    if stacktrace != None:
        token1 = f"jinja2{os.sep}environment.py\", " # Jinja2 hack
        token2 = f"jinja2{os.sep}runtime.py\", " # Jinja2 hack
        startidx = -1
        while True:
            idx = stacktrace.find(token1, startidx+1)
            if idx < 0:
                idx = stacktrace.find(token2, startidx+1)
                if idx < 0:
                    break # Token not found
                startidx = stacktrace.find('\n', idx+len(token2))
            else:
                startidx = stacktrace.find('\n', idx+len(token1))
            if startidx < 0:
                startidx = len(stacktrace)
                break # No eol after token
            if not stacktrace.startswith("    ", startidx+1):
                continue # No inner traceback
            startidx = stacktrace.find('\n', startidx+1)
            if startidx < 0:
                startidx = len(stacktrace)
                break # No eol after inner traceback
        if startidx < 0:
            return exception
        stacktrace = stacktrace[startidx+1:]
        if len(stacktrace) == 0:
            stacktrace = None
        else:
            # Clean autojinja stacktrace
            token = f"autojinja{os.sep}templates.py\", "
            idx = stacktrace.find(token)
            if idx >= 0:
                startidx = stacktrace.rfind('\n', 0, idx)
                if startidx < 0: # No eol before token
                    stacktrace = None
                else:
                    stacktrace = stacktrace[:startidx]
    if message == None:
        message = ""
    message = f"\n{stacktrace}\n{message}" if stacktrace else f"\n{message}"
    return wrap_exception(exception, message)

def clean_traceback(exception):
    """ Removes the traceback of the given exception.
    """
    return prepend_jinja2_traceback(exception).with_traceback(None)

### --debug option enabled

class DebugAttributeError(BaseException):
    def __init__(self, message):
        super().__init__(message)

_disallowed_stacktraces = [
    "autojinja_wrapped_fcall",
    "autojinja_fcall",
    "autojinja_fget",
    "autojinja_fset",
    "autojinja_fdel"
]

def clean_wrapped_stacktrace(exception):
    """ Cleans the the given exception's stacktrace from useless wrapper calls.
    """
    if exception.__class__.__module__ != "builtins":
        fullname = f"{exception.__class__.__module__}.{exception.__class__.__qualname__}: "
    else:
        fullname = f"{exception.__class__.__qualname__}: "
    stacktrace = traceback.format_exc()
    lines = stacktrace.splitlines()
    start = -1
    for i in range(0, len(lines), 1):
        if lines[i].endswith("line 15, in autojinja_wrapped_fcall") \
        or lines[i].endswith("line 18, in autojinja_wrapped_fcall"): # /!\ easily breakable
            start = i
            break
    end = -1
    if start > 0:
        for i in range(len(lines)-1, start, -1):
            if lines[i] == fullname:
                end = i+1
                break
    message = io.StringIO()
    state = 0 # 0, 1
    for line in lines[:start]:
        if state == 1 and line.startswith("    "):
            state = 0 # Skip
        elif line.startswith("Traceback (most recent call last)"):
            pass # Skip
        elif not any(x in line for x in _disallowed_stacktraces):
            message.write(f"\n{line}")
        else:
            state = 1 # Skip
    for line in lines[end:]:
        message.write(f"\n{line}")
    return message.getvalue()

_wrapped_properties = set()

def _autojinja_fcall(wrapped_callable):
    def autojinja_fcall(*args, **kwargs):
        return wrapped_callable(*args, **kwargs)
    return wrap_callable(autojinja_fcall)

def _autojinja_fget(wrapped_property):
    def autojinja_fget(self):
        return wrapped_property.fget(self)
    return wrap_callable(autojinja_fget)

def _autojinja_fset(wrapped_property):
    def autojinja_fset(self, value):
        return wrapped_property.fset(self, value)
    return wrap_callable(autojinja_fset)

def _autojinja_fdel(wrapped_property):
    def autojinja_fdel(self):
        return wrapped_property.fdel(self)
    return wrap_callable(autojinja_fdel)

_disallowed_modules = set([
    "autojinja.defaults",
    "autojinja.exceptions",
    "autojinja.main",
    "autojinja.parser",
    "autojinja.path",
    "autojinja.templates",
    "autojinja.utils",
    "builtins",
])

def wrap_object_attributes(obj):
    """ Wraps all properties and methods of the given objects in order to bypass
        AttributeError handling of Jinja and provide better exception stacktraces.
    """
    if str(obj.__class__.__module__) in _disallowed_modules:
        return obj
    for name in dir(obj):
        try:
            if not hasattr(obj.__class__, name):
                continue
            attribute = getattr(obj.__class__, name)
            if callable(attribute):
                wrapped_callable = getattr(obj, name)
                if str(wrapped_callable.__name__) == "autojinja_wrapped_fcall":
                    continue # Already replaced
                setattr(obj, name, _autojinja_fcall(wrapped_callable))
            elif isinstance(attribute, property):
                wrapped_property = attribute
                key = f"{obj.__class__}.{name}"
                if key in _wrapped_properties:
                    continue # Already replaced
                _wrapped_properties.add(key)
                setattr(obj.__class__, name, property(fget=_autojinja_fget(wrapped_property),
                                                      fset=_autojinja_fset(wrapped_property),
                                                      fdel=_autojinja_fdel(wrapped_property)))
        except:
            pass # Skip attribute
    return obj
